/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Normalization.hh"

#include <Geometry/KMatrix.hh> 
  //#include <Geometry/RMatrix.hh> 
#include <Geometry/HomgPoint2D.hh> 
#include <Math/Matrix4x4.hh> 
#include <Debug/Debug.hh> 

using namespace BIAS;
using namespace std;

int Normalization::
Hartley(const vector<HomgPoint2D>& Points, vector<HomgPoint2D>& NormPoints,
        KMatrix& Normalization) const 
{
  Normalization.SetIdentity();
	const unsigned int uiPointCount(Points.size());
	// we need at least two points for a stddev
	if (uiPointCount<2) return -1;
		
  int result = 0;
  
  NormPoints.reserve(uiPointCount);
#ifdef BIAS_DEBUG
  if (NormPoints.size()>0) {
    BIASERR("non-empty vector supplied !");
    BIASABORT;
  }
#endif
  // compute mean 
  double meanX(0), meanY(0);
  for (unsigned int i=0; i<uiPointCount; i++) {
    meanX += Points[i][0];
    meanY += Points[i][1];
#ifdef BIAS_DEBUG
    if (Points[i][2]!=1.0) {
      BIASERR("Supplied non-homogenized point ("<<i<<") ! "<<Points[i]);
      BIASABORT;
    }
#endif
  }
  meanX /= (double)uiPointCount;
  meanY /= (double)uiPointCount;
  
  // compute mean distance from center
  double meandist(0);
  for (unsigned int i=0; i<uiPointCount; i++) {
    const BIAS::HomgPoint2D& CurPoint = Points[i];
    meandist += sqrt((CurPoint[0]-meanX) * (CurPoint[0]-meanX)  +
                     (CurPoint[1]-meanY) * (CurPoint[1]-meanY));
  }
  meandist /= (double)uiPointCount;

  if (meandist>0.0) {
    // compute Normalization such that mean is zero and mean distance is sqrt2
    Normalization[0][0] = M_SQRT2/meandist;
    Normalization[0][2] = -meanX/meandist;
    Normalization[1][1] = M_SQRT2/meandist;
    Normalization[1][2] = -meanY/meandist;
  } else {
    // stddev is zero, all the same point !!!
    result = -2;
  }
	HomgPoint2D p;
  // now normalize all the points
  for (unsigned int i=0; i<uiPointCount; i++) {
		Normalization.Mult(Points[i], p);
		NormPoints.push_back(p.Homogenize());
	}
	return result;									 
}

#undef M_SQRT3
