/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#ifndef __ROI_hh__
#define __ROI_hh__

#include <Common/BIASpragmaStart.hh>

#include <Debug/Debug.hh>
#include <Debug/Error.hh>
#include <iostream>
#include <vector>
#include <cmath>

#define ROI_VERSION_NUMBER 200

namespace BIAS {

  // forward declaration to avoid mutual inclusion
  class ImageBase;

  enum ERoiType { ROI_Corners, ROI_Mask, ROI_Points, ROI_Rows };

  std::ostream& operator<<(std::ostream& os, const enum ERoiType& type);

  /** @class Position
      @ingroup g_image_base
      @brief stores valid/nvalid positions in image

      @author woelk 07/2005 */
  class BIASImageBase_EXPORT Position 
  {
  public:
    Position() : x(0), y(0) {};
    Position(unsigned xx, unsigned yy) : x(xx), y(yy) {};
    Position(const Position &pos) { *this = pos; } 

    ~Position() {};

    /// assignment operator
    Position& operator=(const Position& p) { x=p.x; y=p.y; return *this; }

    /// comparison, needed for sorting
    bool operator<(const Position& p) const
    { return (y==p.y)?(x<p.x):(y<p.y); }

    friend BIASImageBase_EXPORT 
      std::ostream& operator<<(std::ostream& os, const Position &pos);

    /// member variables
    unsigned x, y;
  };

  /** @class ROI
      @ingroup g_image_base
      @brief class for handling different region of interest (ROI) 
      representations...

      The ROI class usually is member of an ImageBase and is stored together 
      with the ImageBase. If an ImageBase is written to disk in the binary mip 
      format, the ROI member is also automatically written in the same file.

      Flags indication the validity of the single representations exist within 
      the ROI class.

      Right now, three different representations are possible: 
      - \b Corner representation:
      The corner representation represents a rectangular ROI. The corner 
      representation holds upper left and lower right corners defining a 
      rectangular roi region. The upper left corner is included in the roi 
      while the lower right corner is excluded from the roi. When ROI belongs
      to image and no roi is set, upper left corner is (0, 0) and lower right
      corner is given by (width, height). The corner representation is always
      valid.
      - \b Map representations:
      If the ROI belongs to an ImageBase, the MapData() and MapDataArray() 
      must be of the same size as the Image. This is usually nesured by the 
      ImageBase.Init() functions. The MapData() is of Type unsigned char.
      A value of 0 mneans not belonging to ROI, while any other value means
      belonging to ROI.
      - \b Vector representation:
      Every position in the vector belongs to roi, while every position not 
      in the vector does not belong to the ROI. The VectorValid() flag 
      indicates the validity of the vector entries.
      - \b Row representation:
      A start and an endn point ist stored fro each image row. It is thus
      possible to store a convex ROI very memory efficient

      The ROI class can later on be expanded in a way, that the vector does 
      not only hold positions, but maybe somthing like more complex mask 
      patches... 
    
      \todo TODO change ROI memebr in ImageBase to pointer to:
      - share a ROI between multiple images 

      @author woelk 07/2005 redesign woelk 02/2008 (c) www.vision-n.de */
  class BIASImageBase_EXPORT ROI : public Debug
  {
  public:
    /** @name construction and destruction 
        @{ */

    /// destructor 
    ~ROI();

    /// str. constructor
    ROI();

    /// @brief calls operator=
    ROI(const ROI& roi);

    /** Deletes internal memory, sets mask and vector invalid.
        @author woelk 07/2005 */
    void Release();

    /** Reallocates Mask_ and MaskArray_, if necessary. 
        @author woelk 07/2005 */
    void Resize(unsigned width, unsigned height);

    /** @brief Delete region of interest
        @author woelk 07/2005 **/
    void UnsetROI();

    /// assignment operator
    ROI& operator=(const ROI& roi);

    /// alloc and free of internal memory
    void SetROIType(const enum ERoiType& type);

    //@}

    /** @name access function
        @{ */

    /** @brief Sets a rectangular region of interest
        @author woelk 07/2005 */
    int SetCorners(unsigned UpperLeftX, unsigned UpperLeftY, 
                   unsigned LowerRightX, unsigned LowerRightY);

    /** @brief Return the region of interest, by saving the 
        coordinates within the variables defined by the parameters
        @author woelk 07/2005 */    
    inline void GetCorners(unsigned &UpperLeftX, unsigned &UpperLeftY,
                           unsigned &LowerRightX,
                           unsigned &LowerRightY) const;
    /** @brief Return the region of interest, by saving the 
        coordinates within the variables defined by the parameters
        @author woelk 07/2005 */    
    inline void GetCorners(int& UpperLeftX, int& UpperLeftY,
                           int& LowerRightX, int& LowerRightY) const;  

    /** @brief fetch individual ROI corner coordinate

        efficient inline access, used by BackwardMapping in most inner loop
        @author koeser 08/2007 */
    inline unsigned GetCornerUpperLeftX() const { return UpperLeftX_; } 
      
    /** @brief fetch individual ROI corner coordinate

        efficient inline access, used by BackwardMapping in most inner loop
        @author koeser 08/2007 */
    inline unsigned GetCornerUpperLeftY() const { return  UpperLeftY_; }
      
    /** @brief fetch individual ROI corner coordinate

        efficient inline access, used by BackwardMapping in most inner loop
        @author koeser 08/2007 */
    inline unsigned GetCornerLowerRightX() const { return LowerRightX_; }
      
    /** @brief fetch individual ROI corner coordinate

        efficient inline access, used by BackwardMapping in most inner loop
        @author koeser 08/2007 */
    inline unsigned GetCornerLowerRightY() const { return LowerRightY_; }


    /** @brief ROI check if pixel position is inside the ROI
        moved here from image base, does only check bounding box roi by now */
    inline bool IsInROI(const double& x, const double& y) const;

    /** returns an image of StorageType unsigned char, where every 
        pixel not in the ROI is set to UCHAR_MAX and every other pixel to 0
        @author woelk 07/2005 */
    int GetMaskImage(ImageBase& im) const;

    /** every pixel with value!=0.0 is in the ROI, every pixel with
        value==0.0 is not in the ROI 
        @author woelk 07/2005 */
    int SetMaskImage(const ImageBase& im);

    /** Direct access to the mask data
        @author woelk 07/2005 */
    void SetMask(const unsigned& x, const unsigned& y, const bool val);
    
    /** Direct access to the mask data, const version
        @author woelk 07/2005 */
    bool Mask(const unsigned& x, const unsigned& y) const;

    /** Sets MaskValid_=false and VectorValid_=true.
        Copies pos -> Vector_.
        Sorts Vector_.
        @author woelk 07/2005 */
    void SetVector(std::vector<Position>& pos);

    /// returns a pointer to the vector representation of the ROI
    inline std::vector<Position>* GetVector()
    { if (RoiType_!=ROI_Points) return NULL; return &Vector_; }

    /** get a copy of mask vector. Returns true if vector is valid
        @author woelk 07/2005 */
    bool GetVector(std::vector<Position>& vec) const;

    /// width capacity of roi
    inline unsigned GetWidth() const
    { return Width_; }

    /// height capacity of roi
    inline unsigned GetHeight() const
    { return Height_; }

    inline bool GetRows(std::vector<unsigned>& start, 
                        std::vector<unsigned>& end) const
    { start = RowStart_; end = RowEnd_; return (RoiType_==ROI_Rows); }

    void SetRows(const std::vector<unsigned>& start, 
                 const std::vector<unsigned>& end);

    //@}

    /** @name checking
        @{ */

    /// is the mask image valid?
    inline enum ERoiType GetROIType() const
    { return RoiType_; }

    /// is the position vector valid?
    inline bool MaskValid() const
    { return (RoiType_==ROI_Mask); }

    /// is the position vector valid?
    inline bool VectorValid() const
    { return (RoiType_==ROI_Points); }

    /** Returns true if both ROIs have the same size.
        Only the rectangular ROI as defined by the corners is validated.
        @author woelk 07/2005 */
    bool SameSize(const ROI& roi) const;

    //@}

    /** @name io functions
        @{ */

    /// 
    friend BIASImageBase_EXPORT std::ostream& 
      operator<<(std::ostream& os, const ROI& roi);

    /** binary output to stream
        @author woelk 07/2005 */
    int WriteBinary(std::ostream& os) const;

    /** binary input from stream
        @author woelk 07/2005 */
    int ReadBinary(std::istream& is);

    //@}

    /** @name conversion functions
        @{ */

    /** Fills vector representation of ROI from mask representation.
        Returns 0 if mask is valid. Sets VectorValid flag if maks is valid. 
        @author woelk 07/2005 */ 
    int Mask2Vector();

    /** Fills vector representation of ROI from mask representation.
        Returns 0 if vector is valid. Sets MaskValid flag if vector is valid. 
        @author woelk 07/2005  */ 
    int Vector2Mask();

    /** Fills corner/bounding box representation of ROI from vector 
        representation. Returns 0 if vector is valid.  
        @author woelk 07/2005 */ 
    int Vector2Corners();

    /** Fills corner/bounding box representation of ROI from mask 
        representation. Returns 0 if mask is valid.  */ 
    int Mask2Corners();

    /** Fills mask representation of ROI from corner representation */ 
    void Corners2Mask();

    /** Fills vector representation of ROI from corner representation */ 
    void Corners2Vector();

    //@}

  protected:
    /// which method is used to store the ROI
    enum ERoiType RoiType_;
    /// maximum dimension of roi data
    unsigned Width_, Height_;

    /// internal data for ROI of type ROI_Corners
    /// rectangular region of interest defined by upper left and lower right
    /// corners. The upperleft corner is part of the rectangle, while the
    /// lower is just outside
    unsigned UpperLeftX_, UpperLeftY_; 
    unsigned LowerRightX_, LowerRightY_; 

    /// internal data for ROI of type ROI_Mask
    /// the mask data: Mask_[y*Width_+x] corresponds to pixel at (x,y)
    /// a bool vector is used, because the bool std::vector is internally 
    /// stored as a bitfield
    std::vector<bool> Mask_;

    /// internal data for ROI of type ROI_Points
    /// Vector instance of ROI, always sorted from top left to bottom right
    std::vector<Position> Vector_; 

    /// internal data for ROI of type ROI_Rows
    /// the ROI is specified by a min and a max value for each row in the image
    std::vector<unsigned> RowStart_, RowEnd_;
    
    /// version number of ROI class
    static const int Version_;

  }; // class

  /** human readable output operator for ROI, mainly for debugging 
      @author woelk 07/2005 */ 
  BIASImageBase_EXPORT std::ostream& 
  operator<<(std::ostream& os, const ROI& roi);


  ///////////////////////////////////////////////////////////////////////
  // inline functions
  ///////////////////////////////////////////////////////////////////////

  inline
  void ROI::GetCorners(unsigned &UpperLeftX, unsigned &UpperLeftY,
                       unsigned &LowerRightX,unsigned &LowerRightY) const
  {
    UpperLeftX  = UpperLeftX_;
    UpperLeftY  = UpperLeftY_;
    LowerRightX = LowerRightX_;
    LowerRightY = LowerRightY_;
  }

  inline
  void ROI::GetCorners(int& UpperLeftX, int& UpperLeftY,
                       int& LowerRightX,  int& LowerRightY) const 
  {
    UpperLeftX  = (int)UpperLeftX_;
    UpperLeftY  = (int)UpperLeftY_;
    LowerRightX = (int)LowerRightX_;
    LowerRightY = (int)LowerRightY_;
  }

  inline bool ROI::IsInROI(const double& x, const double& y) const
  {
    switch (RoiType_){
    case ROI_Corners:
      return ((UpperLeftX_ <= x) &&  ((LowerRightX_-1)>=x) &&
              (UpperLeftY_ <= y) &&  ((LowerRightY_-1)>=y));
      break;
    case ROI_Mask:
       {
        unsigned iy = (unsigned)rint(y);
        unsigned ix = (unsigned)rint(x);
        if (ix>=Width_ || iy>=Height_) return false;
        return Mask(ix, iy);
       }
      break;
    case ROI_Points:
      {
        unsigned iy = (unsigned)rint(y);
        unsigned ix = (unsigned)rint(x);
        std::vector<Position>::const_iterator it;
        for (it=Vector_.begin(); it!=Vector_.end(); it++){
          if (it->x==ix && it->y==iy) return true;
        }
        return false;
      }
      break;
    case ROI_Rows:
      {
        int iy = (int)rint(y);
        if (iy<0 || iy>=(int)Height_) return false;
        int ix = (int)rint(x);
        return (ix>=(int)RowStart_[iy] && ix<(int)RowEnd_[iy]);
      }
      break;
    }
    return false;
  }

} // namespace



#include <Common/BIASpragmaEnd.hh>

#endif // __ROI_hh__
